@@ -1,9 +1,10 @@
+import copy
 import hashlib
 import logging
+import multiprocessing
 import pprint
 import random
 import time
-from copy import copy
 
 from django.db import DEFAULT_DB_ALIAS, DatabaseError
 from django.db.backends.mysql import base
@@ -14,8 +15,8 @@ LOGGER = logging.getLogger(__name__)
 class NodeState:
     RETRY_INTERVAL = 30
 
-    def __init__(self):
-        self.nodes = dict()
+    def __init__(self, store):
+        self.nodes = store
 
     def add_nodes(self, nodes):
         for node in nodes:
@@ -34,7 +35,7 @@ class NodeState:
         return (x for x, y in self.nodes.items() if y is None or time.time() > y + self.RETRY_INTERVAL)
 
 
-NODE_STATE = NodeState()
+NODE_STATE = NodeState(multiprocessing.Manager().dict())
 
 
 class CursorWrapper:
@@ -47,9 +48,9 @@ class CursorWrapper:
     @property
     def cursor(self):
         if self._cursor is None:
+            self._backend.failover_history.append([])
             if self._primary:
                 self._cursor = self._backend.create_primary_cursor()
-                self._backend.failover_history.append([])
             else:
                 self._cursor = self._backend.create_secondary_cursor()
         return self._cursor
@@ -92,49 +93,43 @@ class CursorWrapper:
             self._cursor = None
 
     def _failover_cursor(self, item):
-        # clear and disable query history if size exceeds limit for this connection
-        if self._backend.failover_enable:
-            if self._backend.failover_history_size >= self._backend.failover_history_limit:
-                LOGGER.info('Failover history limit reached. Disabling history for this connection.')
-                self._backend.failover_history.clear()
-                self._backend.failover_history_size = 0
-                self._backend.failover_history_enable = False
+        if not self._backend.failover_active:
+            return getattr(self.cursor, item)
 
         try:
             obj = getattr(self.cursor, item)
-        except base.Database.OperationalError as e:
+        except DatabaseError as e:
             self._handle_exc(e)
             obj = getattr(self._cursor, item)
 
-        if item != '_executed' and self._primary and self._backend.failover_enable:
-            def wrap(func):
-                def decor(*args, **kwargs):
-                    try:
-                        ret = func(*args, **kwargs)
-                    except base.Database.OperationalError as e:
-                        self._handle_exc(e)
-                        ret = getattr(self._cursor, func.__name__)(*args, **kwargs)
-                    if self._backend.failover_history:
-                        self._backend.failover_history[-1].append((
-                            func.__name__,
-                            args,
-                            kwargs,
-                            hashlib.sha1(pprint.pformat(ret).encode()).hexdigest()
-                        ))
-                        self._backend.failover_history_size += 1
-                    return ret
-                return decor
-            if hasattr(obj, '__call__'):
-                obj = wrap(obj)
-            else:
+        def wrap(func):
+            def decor(*args, **kwargs):
+                try:
+                    ret = func(*args, **kwargs)
+                except DatabaseError as exc:
+                    self._handle_exc(exc)
+                    ret = getattr(self._cursor, func.__name__)(*args, **kwargs)
                 if self._backend.failover_history:
                     self._backend.failover_history[-1].append((
-                        item,
-                        None,
-                        None,
-                        hashlib.sha1(pprint.pformat(obj).encode()).hexdigest()
+                        func.__name__,
+                        args,
+                        kwargs,
+                        hashlib.sha1(pprint.pformat(ret).encode()).hexdigest()
                     ))
                     self._backend.failover_history_size += 1
+                return ret
+            return decor
+        if hasattr(obj, '__call__'):
+            obj = wrap(obj)
+        else:
+            if self._backend.failover_history:
+                self._backend.failover_history[-1].append((
+                    item,
+                    None,
+                    None,
+                    hashlib.sha1(pprint.pformat(obj).encode()).hexdigest()
+                ))
+                self._backend.failover_history_size += 1
 
         return obj
 
@@ -145,11 +140,14 @@ class CursorWrapper:
         if self._in_handle_exc:
             raise exc
         self._in_handle_exc = True
-        if self._backend.failover_enable:
+        if self._backend.failover_active:
             error_code = exc.args[0]
-            if error_code == 2006: # server has gone away
+            if error_code in (
+                    2006,  # MySQL server has gone away
+                    2013,  # Lost connection to MySQL server during query
+            ):
                 autocommit = self._backend.autocommit
-                history = copy(self._backend.failover_history)
+                history = copy.copy(self._backend.failover_history)
                 history_size = self._backend.failover_history_size
                 # try to gracefully close the original cursor and connection even if it will most certainly fail
                 try:
@@ -166,13 +164,7 @@ class CursorWrapper:
                 self._backend.set_autocommit(autocommit)
                 self._backend.failover_history = history
                 self._backend.failover_history_size = history_size
-                try:
-                    raise exc
-                except Exception as e:
-                    logging.getLogger('django').error(
-                        'Replaying %d cursors from failover history after %s' % (len(history), str(e)),
-                        exc_info=True
-                    )
+                LOGGER.warning('Replaying %d cursors from failover history after %s' % (len(history), str(exc)))
                 self._cursor = self._backend.replay_history()
                 self._in_handle_exc = False
                 return
@@ -182,31 +174,33 @@ class CursorWrapper:
 
 class DatabaseWrapper(base.DatabaseWrapper):
     base_settings = None
+    failover_history = list()
+    failover_history_size = 0
     in_write_transaction = False
     primary_connected = False
     secondary_synced = True
+
+    _failover_active = None
+    _failover_enable = None
     _secondary_wrapper = None
 
     def __init__(self, settings_dict, alias=DEFAULT_DB_ALIAS):
-        self.base_settings = settings_dict.copy()
-        NODE_STATE.add_nodes(settings_dict.get('NODES', ()))
-        if 'OPTIONS' not in settings_dict:
-            settings_dict['OPTIONS'] = dict()
-        settings_dict['OPTIONS'].pop('unix_socket', None)
-        self.failover_enable = settings_dict['OPTIONS'].pop('failover_enable', False)
-        self.failover_history_limit = settings_dict['OPTIONS'].pop('failover_history_limit', 1000)
-        self.failover_history = list()
-        self.failover_history_size = 0
-        self.wsrep_sync_after_write = settings_dict['OPTIONS'].pop('wsrep_sync_after_write', False)
-        super(DatabaseWrapper, self).__init__(settings_dict, alias=alias)
+        self.base_settings = copy.deepcopy(settings_dict)
+        NODE_STATE.add_nodes(self.base_settings.get('NODES', ()))
+        if 'OPTIONS' not in self.base_settings:
+            self.base_settings['OPTIONS'] = dict()
+        self.base_settings['OPTIONS'].pop('unix_socket', None)
+        self.failover_enable = self.base_settings['OPTIONS'].pop('failover_enable', False)
+        self.failover_history_limit = self.base_settings['OPTIONS'].pop('failover_history_limit', 1000)
+        self.wsrep_sync_after_write = self.base_settings['OPTIONS'].pop('wsrep_sync_after_write', False)
+        super(DatabaseWrapper, self).__init__(self.base_settings, alias=alias)
 
     def close(self):
         if self._secondary_wrapper is not None:
             self._secondary_wrapper.close()
             self._secondary_wrapper = None
         super(DatabaseWrapper, self).close()
-        self.failover_history.clear()
-        self.failover_history_size = 0
+        self.failover_history_reset()
 
     def connect(self):
         self.connect_to_node(primary=True)
@@ -225,7 +219,7 @@ class DatabaseWrapper(base.DatabaseWrapper):
                 nodes.insert(0, preferred_host)
         for node in nodes:
             LOGGER.info('connect %s to node %s' % ('primary' if primary else 'secondary', node))
-            settings_dict = self.base_settings.copy()
+            settings_dict = copy.deepcopy(self.base_settings)
             settings_dict['ENGINE'] = 'django.db.backends.mysql'
             for k, v in settings_dict['NODES'].get(node, {}).items():
                 settings_dict[k] = v
@@ -262,8 +256,7 @@ class DatabaseWrapper(base.DatabaseWrapper):
             return super(DatabaseWrapper, self).create_cursor(name=name)
         else:
             if self.autocommit:
-                self.failover_history.clear()
-                self.failover_history_size = 0
+                self.failover_history_reset()
             return CursorWrapper(self)
 
     def create_primary_cursor(self):
@@ -274,6 +267,31 @@ class DatabaseWrapper(base.DatabaseWrapper):
         cursor = self.secondary_wrapper.cursor()
         return base.CursorWrapper(cursor)
 
+    @property
+    def failover_active(self):
+        if self._failover_active and self.failover_history_size >= self.failover_history_limit:
+            self.failover_active = False
+        return self._failover_active
+
+    @failover_active.setter
+    def failover_active(self, value):
+        if not value:
+            self.failover_history_reset()
+        self._failover_active = value
+
+    @property
+    def failover_enable(self):
+        return self._failover_enable
+
+    @failover_enable.setter
+    def failover_enable(self, value):
+        self.failover_active = value
+        self._failover_enable = value
+
+    def failover_history_reset(self):
+        self.failover_history.clear()
+        self.failover_history_size = 0
+
     @property
     def secondary_wrapper(self):
         if self._secondary_wrapper is None:
@@ -283,8 +301,9 @@ class DatabaseWrapper(base.DatabaseWrapper):
     def _set_autocommit(self, autocommit):
         if autocommit:
             self.in_write_transaction = False
-            self.failover_history.clear()
-            self.failover_history_size = 0
+        if autocommit != self.autocommit:
+            self.failover_history_reset()
+            self.failover_active = self.failover_enable
         return super(DatabaseWrapper, self)._set_autocommit(autocommit)
 
     def sync_wait_secondary(self):
@@ -294,11 +313,13 @@ class DatabaseWrapper(base.DatabaseWrapper):
             result = cursor.fetchone()
             cursor.close()
             primary_gtid = result[0].decode('utf-8')
-            if primary_gtid != '00000000-0000-0000-0000-000000000000:-1':
-                cursor = self.secondary_wrapper.connection.cursor()
+            cursor = self.secondary_wrapper.connection.cursor()
+            try:
                 cursor.execute('SELECT WSREP_SYNC_WAIT_UPTO_GTID(%s)', (primary_gtid,))
-                cursor.close()
                 LOGGER.debug('Secondary sync upto %s' % primary_gtid)
+            except base.Database.OperationalError as e:
+                LOGGER.warning('Could not sync secondary upto %s: %s' % (primary_gtid, str(e)))
+            cursor.close()
         self.secondary_synced = True
 
     def replay_history(self):
@@ -314,7 +335,7 @@ class DatabaseWrapper(base.DatabaseWrapper):
                 else:
                     result = attr(*args, **kwargs)
                 if check != hashlib.sha1(pprint.pformat(result).encode()).hexdigest():
-                    raise Exception('Replay checksum does not match')
+                    raise DatabaseError('Replay checksum does not match')
 
             # do not close the cursor if this is the last history entry
             # the cursor should then have a comparable state as the original cursor
